@@ -1,5 +1,6 @@
 #!/usr/bin/env python3
 
+import asyncio
 import ipaddress
 import hashlib
 import json
@@ -8,6 +9,7 @@ import os
 import pickle
 import re
 import socket
+import time
 import unicodedata
 
 from base64 import b64decode, b64encode
@@ -22,6 +24,7 @@ from defang import refang  # type: ignore
 from playwrightcapture import Capture, PlaywrightCaptureException
 from playwrightcapture.capture import CaptureResponse as PlaywrightCaptureResponse
 from redis import Redis
+from redis.exceptions import ConnectionError as RedisConnectionError
 from ua_parser import user_agent_parser  # type: ignore
 
 BROWSER = Literal['chromium', 'firefox', 'webkit']
@@ -53,8 +56,13 @@ class CaptureError(LacusCoreException):
     pass
 
 
+class RetryCapture(LacusCoreException):
+    pass
+
+
 @unique
 class CaptureStatus(IntEnum):
+    '''The status of the capture'''
     UNKNOWN = -1
     QUEUED = 0
     DONE = 1
@@ -62,6 +70,7 @@ class CaptureStatus(IntEnum):
 
 
 class CaptureResponse(TypedDict, total=False):
+    '''A capture made by Lacus. With the base64 encoded image and downloaded file decoded to bytes.'''
 
     status: int
     last_redirected_url: str
@@ -76,6 +85,7 @@ class CaptureResponse(TypedDict, total=False):
 
 
 class CaptureResponseJson(TypedDict, total=False):
+    '''A capture made by Lacus. With the base64 encoded image and downloaded file *not* decoded.'''
 
     status: int
     last_redirected_url: Optional[str]
@@ -90,6 +100,7 @@ class CaptureResponseJson(TypedDict, total=False):
 
 
 class CaptureSettings(TypedDict, total=False):
+    '''The capture settings that can be passed to Lacus.'''
 
     url: Optional[str]
     document_name: Optional[str]
@@ -104,8 +115,11 @@ class CaptureSettings(TypedDict, total=False):
     http_credentials: Optional[Dict[str, int]]
     viewport: Optional[Dict[str, int]]
     referer: Optional[str]
+    force: Optional[bool]
+    recapture_interval: Optional[int]
+    priority: Optional[int]
 
-    depth: Optional[int]
+    depth: int
     rendered_hostname_only: bool  # Note: only used if depth is > 0
 
 
@@ -115,17 +129,26 @@ def _json_encode(obj: Union[bytes]) -> str:
 
 
 class LacusCore():
+    """Capture URLs or web enabled documents using PlaywrightCapture.
 
-    def __init__(self, redis_connector: Redis, tor_proxy: Optional[str]=None, only_global_lookups: bool=True) -> None:
+    :param redis_connector: Pre-configured connector to a redis instance.
+    :param tor_proxy: URL to a SOCKS 5 tor proxy. If you have tor installed, this is the default: socks5://127.0.0.1:9050.
+    :param only_global_lookups: Discard captures that point to non-public IPs.
+    :param max_retries: How many times should we re-try a capture if it failed.
+    """
+
+    def __init__(self, redis_connector: Redis, tor_proxy: Optional[str]=None, only_global_lookups: bool=True, max_retries: int=3) -> None:
         self.logger = logging.getLogger(f'{self.__class__.__name__}')
         self.logger.setLevel('INFO')
 
         self.redis = redis_connector
         self.tor_proxy = tor_proxy
         self.only_global_lookups = only_global_lookups
+        self.max_retries = max_retries
 
-    def check_redis_up(self):
-        return self.redis.ping()
+    def check_redis_up(self) -> bool:
+        """Check if redis is reachable"""
+        return bool(self.redis.ping())
 
     @overload
     def enqueue(self, *, settings: Optional[CaptureSettings]=None) -> str:
@@ -150,7 +173,7 @@ class LacusCore():
                 recapture_interval: int=300,
                 priority: int=0
                 ) -> str:
-        pass
+        ...
 
     def enqueue(self, *,
                 settings: Optional[CaptureSettings]=None,
@@ -171,8 +194,39 @@ class LacusCore():
                 recapture_interval: int=300,
                 priority: int=0
                 ) -> str:
+        """Enqueue settings.
+
+        :param settings: Settings as a dictionary
+
+        :param url: URL to capture (incompatible with document and document_name)
+        :param document_name: Filename of the document to capture (required if document is used)
+        :param document: Document to capture itself (requires a document_name)
+        :param depth: [Dangerous] Depth of the capture. If > 0, the URLs of the rendered document will be extracted and captured. It can take a very long time.
+        :param browser: The prowser to use for the capture
+        :param device_name: The name of the device, must be something Playwright knows
+        :param user_agent: The user agent the browser will use for the capture
+        :param proxy: SOCKS5 proxy to use for capturing
+        :param general_timeout_in_sec: The capture will raise a timeout it it takes more than that time
+        :param cookies: A list of cookies
+        :param headers: The headers to pass to the capture
+        :param http_credentials: HTTP Credentials to pass to the capture
+        :param viewport: The viewport of the browser used for capturing
+        :param referer: The referer URL for the capture
+        :param rendered_hostname_only: If depth > 0: only capture URLs with the same hostname as the rendered page
+        :param force: Force recapture, even if the same one was already done within the recapture_interval
+        :param recapture_interval: The time the enqueued settings are kept in memory to avoid duplicates
+        :param priority: The priority of the capture
+
+        :return: UUID, reference to the capture for later use
+        """
         to_enqueue: CaptureSettings
         if settings:
+            if settings.get('force') is not None:
+                force = settings.pop('force')  # type: ignore
+            if settings.get('recapture_interval') is not None:
+                recapture_interval = settings.pop('recapture_interval')  # type: ignore
+            if settings.get('priority') is not None:
+                priority = settings.pop('priority')  # type: ignore
             to_enqueue = settings
         else:
             to_enqueue = {'depth': depth, 'rendered_hostname_only': rendered_hostname_only}
@@ -181,8 +235,6 @@ class LacusCore():
             elif document_name and document:
                 to_enqueue['document_name'] = _secure_filename(document_name)
                 to_enqueue['document'] = document
-            else:
-                raise Exception('Needs either a URL or a document_name *and* a document.')
             if browser:
                 to_enqueue['browser'] = browser
             if device_name:
@@ -204,8 +256,8 @@ class LacusCore():
             if referer:
                 to_enqueue['referer'] = referer
 
+        hash_query = hashlib.sha512(pickle.dumps(to_enqueue)).hexdigest()
         if not force:
-            hash_query = hashlib.sha512(pickle.dumps(to_enqueue)).hexdigest()
             if (existing_uuid := self.redis.get(f'lacus:query_hash:{hash_query}')):
                 if isinstance(existing_uuid, bytes):
                     return existing_uuid.decode()
@@ -249,10 +301,17 @@ class LacusCore():
         ...
 
     def get_capture(self, uuid: str, *, decode: bool=False) -> Union[CaptureResponse, CaptureResponseJson]:
+        """Get the results of a capture, in a json compatible format or not
+
+        :param uuid: The UUID if the capture (given by enqueue)
+        :param decode: Decode the capture result or not.
+
+        :return: The capture, decoded or not.
+        """
         to_return: CaptureResponseJson = {'status': CaptureStatus.UNKNOWN}
         if self.redis.zscore('lacus:to_capture', uuid):
             to_return['status'] = CaptureStatus.QUEUED
-        elif self.redis.sismember('lacus:ongoing', uuid):
+        elif self.redis.zscore('lacus:ongoing', uuid) is not None:
             to_return['status'] = CaptureStatus.ONGOING
         elif response := self.redis.get(f'lacus:capture_results:{uuid}'):
             to_return['status'] = CaptureStatus.DONE
@@ -262,15 +321,25 @@ class LacusCore():
         return to_return
 
     def get_capture_status(self, uuid: str) -> CaptureStatus:
+        """Get the status of a capture
+
+        :param uuid: The UUID if the capture (given by enqueue)
+
+        :return: The status
+        """
         if self.redis.zscore('lacus:to_capture', uuid):
             return CaptureStatus.QUEUED
-        elif self.redis.sismember('lacus:ongoing', uuid):
+        elif self.redis.zscore('lacus:ongoing', uuid) is not None:
             return CaptureStatus.ONGOING
         elif self.redis.exists(f'lacus:capture_results:{uuid}'):
             return CaptureStatus.DONE
         return CaptureStatus.UNKNOWN
 
     async def consume_queue(self) -> Optional[str]:
+        """Trigger the capture with the highest priority
+
+        :return: The UUID of the capture.
+        """
         value: List[Tuple[bytes, float]] = self.redis.zpopmax('lacus:to_capture')
         if not value or not value[0]:
             return None
@@ -278,68 +347,124 @@ class LacusCore():
         await self.capture(uuid)
         return uuid
 
-    async def capture(self, uuid: str):
-        if self.redis.sismember('lacus:ongoing', uuid):
+    async def capture(self, uuid: str, score: Optional[Union[float, int]]=None):
+        """Trigger a specific capture
+
+        :param uuid: The UUID if the capture (given by enqueue)
+        :param score: Only for internal use, will decide ont he priority of the capture if the try now fails.
+        """
+        if self.redis.zscore('lacus:ongoing', uuid) is not None:
             # the capture is ongoing
             return
 
-        p = self.redis.pipeline()
-        p.zrem('lacus:to_capture', uuid)  # In case the capture is triggered manually
-        p.sadd('lacus:ongoing', uuid)
-        p.execute()
+        retry = False
+        current_score: float = 0.0
+        if score is not None:
+            if isinstance(score, int):
+                current_score = float(score)
+            else:
+                current_score = score
+        if (s := self.redis.zscore('lacus:to_capture', uuid)) is not None:
+            # In case the capture is triggered manually
+            self.redis.zrem('lacus:to_capture', uuid)
+            current_score = s
+
+        self.redis.zadd('lacus:ongoing', {uuid: time.time()})
         try:
             setting_keys = ['depth', 'rendered_hostname_only', 'url', 'document_name',
                             'document', 'browser', 'device_name', 'user_agent', 'proxy',
                             'general_timeout_in_sec', 'cookies', 'headers', 'http_credentials',
                             'viewport', 'referer']
-            result: PlaywrightCaptureResponse
-            to_capture = {}
-            for k, v in zip(setting_keys, self.redis.hmget(f'lacus:capture_settings:{uuid}', setting_keys)):
-                if v is not None:
-                    to_capture[k] = v if k in ['document'] else v.decode()  # Do not decode the document
+            result: PlaywrightCaptureResponse = {}
+            to_capture: CaptureSettings = {}
+            document_as_bytes = b''
+            try:
+                for k, v in zip(setting_keys, self.redis.hmget(f'lacus:capture_settings:{uuid}', setting_keys)):
+                    if v is None:
+                        continue
+                    if k in ['url', 'document_name', 'browser', 'device_name', 'user_agent', 'referer']:
+                        # string
+                        to_capture[k] = v.decode()  # type: ignore
+                    elif k in ['cookies', 'http_credentials', 'viewport']:
+                        # dicts or list
+                        to_capture[k] = json.loads(v)  # type: ignore
+                    elif k in ['proxy', 'headers']:
+                        # can be dict or str
+                        if v[0] == b'{':
+                            to_capture[k] = json.loads(v)  # type: ignore
+                        else:
+                            to_capture[k] = v.decode()  # type: ignore
+                    elif k in ['general_timeout_in_sec', 'depth']:
+                        # int
+                        to_capture[k] = int(v)  # type: ignore
+                    elif k in ['rendered_hostname_only']:
+                        # bool
+                        to_capture[k] = bool(int(v))  # type: ignore
+                    elif k == 'document':
+                        document_as_bytes = b64decode(v)
+                    else:
+                        raise LacusCoreException(f'Unexpected setting: {k}: {v}')
+            except LacusCoreException as e:
+                raise e
+            except Exception as e:
+                raise LacusCoreException(f'Error while preparing settings: {e}')
+
             if not to_capture:
                 result = {'error': f'No capture settings for {uuid}.'}
                 raise CaptureError
 
-            if to_capture.get('document'):
+            url: str = ''
+            if document_as_bytes:
                 # we do not have a URL yet.
-                document_name = Path(to_capture['document_name']).name
+                name = to_capture.pop('document_name', None)
+                if not name:
+                    raise LacusCoreException('No document name provided, settings are invalid')
+                document_name = Path(name).name
                 tmp_f = NamedTemporaryFile(suffix=document_name, delete=False)
                 with open(tmp_f.name, "wb") as f:
-                    f.write(to_capture['document'])
-                url: str = f'file://{tmp_f.name}'
+                    f.write(document_as_bytes)
+                url = f'file://{tmp_f.name}'
             elif to_capture.get('url'):
-                url = to_capture['url'].strip()
+                url = to_capture['url'].strip()  # type: ignore
                 url = refang(url)  # In case we get a defanged url at this stage.
                 if not url.startswith('data') and not url.startswith('http') and not url.startswith('file'):
                     url = f'http://{url}'
             else:
-                result = {'error': f'No valid URL to capture for {uuid}.'}
+                result = {'error': f'No valid URL to capture for {uuid} - {to_capture}'}
                 raise CaptureError
 
             splitted_url = urlsplit(url)
-            if self.only_global_lookups and splitted_url.scheme not in ['data', 'file']:
+            proxy = to_capture.get('proxy')
+            if self.tor_proxy:
+                # check if onion or forced
+                if (proxy == 'force_tor'
+                        or (not proxy
+                            and splitted_url.netloc
+                            and splitted_url.hostname
+                            and splitted_url.hostname.split('.')[-1] == 'onion')):
+                    proxy = self.tor_proxy
+
+            if self.only_global_lookups and not proxy and splitted_url.scheme not in ['data', 'file']:
+                # not relevant if we also have a proxy, or the thing to capture is a data URI or a file on disk
                 if splitted_url.netloc:
                     if splitted_url.hostname and splitted_url.hostname.split('.')[-1] != 'onion':
                         try:
-                            ip = socket.gethostbyname(splitted_url.hostname)
+                            ips_info = socket.getaddrinfo(splitted_url.hostname, None, proto=socket.IPPROTO_TCP)
                         except socket.gaierror:
-                            self.logger.info('Name or service not known')
+                            self.logger.info(f'Unable to resolve {splitted_url.hostname}.')
                             result = {'error': f'Unable to resolve {splitted_url.hostname}.'}
+                            raise RetryCapture
+                        except Exception as e:
+                            result = {'error': f'Issue with hostname resolution ({splitted_url.hostname}): {e}.'}
                             raise CaptureError
-                        if not ipaddress.ip_address(ip).is_global:
-                            result = {'error': 'Capturing ressources on private IPs is disabled.'}
-                            raise CaptureError
+                        for info in ips_info:
+                            if not ipaddress.ip_address(info[-1][0]).is_global:
+                                result = {'error': f'Capturing ressources on private IPs {info[-1][0]} is disabled.'}
+                                raise CaptureError
                 else:
-                    result = {'error': 'Unable to find hostname or IP in the query.'}
+                    result = {'error': f'Unable to find hostname or IP in the query: {url}.'}
                     raise CaptureError
 
-            proxy = to_capture.get('proxy')
-            # check if onion
-            if (not proxy and splitted_url.netloc and splitted_url.hostname
-                    and splitted_url.hostname.split('.')[-1] == 'onion'):
-                proxy = self.tor_proxy
-
             browser_engine: BROWSER = "chromium"
             if to_capture.get('user_agent'):
                 parsed_string = user_agent_parser.ParseUserAgent(to_capture.get('user_agent'))
@@ -355,15 +480,18 @@ class LacusCore():
                 self.logger.info(f'Capturing {url}')
                 async with Capture(browser=browser_engine,
                                    device_name=to_capture.get('device_name'),
-                                   proxy=proxy) as capture:
+                                   proxy=proxy,
+                                   general_timeout_in_sec=to_capture.get('general_timeout_in_sec')) as capture:
                     # required by Mypy: https://github.com/python/mypy/issues/3004
                     capture.headers = to_capture.get('headers')  # type: ignore
-                    if to_capture.get('cookies'):
-                        capture.cookies = json.loads(to_capture.get('cookies'))  # type: ignore
-                    capture.viewport = to_capture.get('viewport')
+                    capture.cookies = to_capture.get('cookies')  # type: ignore
+                    capture.viewport = to_capture.get('viewport')  # type: ignore
                     capture.user_agent = to_capture.get('user_agent')  # type: ignore
                     await capture.initialize_context()
-                    result = await capture.capture_page(url, referer=to_capture.get('referer'))
+                    result = await capture.capture_page(
+                        url, referer=to_capture.get('referer'),
+                        depth=to_capture.get('depth', 0),
+                        rendered_hostname_only=to_capture.get('rendered_hostname_only', True))
             except PlaywrightCaptureException as e:
                 self.logger.exception(f'Invalid parameters for the capture of {url} - {e}')
                 result = {'error': 'Invalid parameters for the capture of {url} - {e}'}
@@ -373,8 +501,28 @@ class LacusCore():
                 result = {'error': f'Something went poorly {url} - {e}'}
                 raise CaptureError
 
+            if hasattr(capture, 'should_retry') and capture.should_retry:
+                # PlaywrightCapture considers this capture elligible for a retry
+                self.logger.info(f'PlaywrightCapture considers {uuid} elligible for a retry.')
+                raise RetryCapture
+        except RetryCapture:
+            # Check if we already re-tried this capture
+            if (current_retry := self.redis.get(f'lacus:capture_retry:{uuid}')) is None:
+                self.redis.setex(f'lacus:capture_retry:{uuid}', 300, self.max_retries)
+            else:
+                self.redis.decr(f'lacus:capture_retry:{uuid}')
+            if current_retry is None or int(current_retry.decode()) > 0:
+                # Just wait a little bit before retrying, expecially if it is the only capture in the queue
+                await asyncio.sleep(5)
+                retry = True
         except CaptureError:
-            self.logger.warning(f'Unable to capture {url} - {uuid}: {result["error"]}')
+            if not result:
+                result = {'error': "No result key, shouldn't happen"}
+                self.logger.exception(f'Unable to capture {uuid}: {result["error"]}')
+            if url:
+                self.logger.warning(f'Unable to capture {url} - {uuid}: {result["error"]}')
+            else:
+                self.logger.warning(f'Unable to capture {uuid}: {result["error"]}')
         else:
             self.logger.info(f'Successfully captured {url} - {uuid}')
         finally:
@@ -382,8 +530,21 @@ class LacusCore():
             if to_capture.get('document'):
                 os.unlink(tmp_f.name)
 
-            p = self.redis.pipeline()
-            p.setex(f'lacus:capture_results:{uuid}', 36000, json.dumps(result, default=_json_encode))
-            p.delete(f'lacus:capture_settings:{uuid}')
-            p.srem('lacus:ongoing', uuid)
-            p.execute()
+            retry_redis_error = 3
+            while retry_redis_error > 0:
+                try:
+                    p = self.redis.pipeline()
+                    if retry:
+                        p.zadd('lacus:to_capture', {uuid: current_score - 1})
+                    else:
+                        p.setex(f'lacus:capture_results:{uuid}', 36000, json.dumps(result, default=_json_encode))
+                        p.delete(f'lacus:capture_settings:{uuid}')
+                    p.zrem('lacus:ongoing', uuid)
+                    p.execute()
+                    break
+                except RedisConnectionError as e:
+                    self.logger.warning(f'Redis Connection Error: {e}')
+                    retry_redis_error -= 1
+                    await asyncio.sleep(5)
+            else:
+                self.logger.critical(f'Unable to connect to redis and to push the result of the capture {uuid}.')
